import unittest
import pandas as pd
import numpy as np
from recsyslearn.metrics import Entropy, KullbackLeibler, MutualInformation, Novelty, Coverage
from tests.utils import first_example, second_example, item_groups, user_groups, rel_matrix_1, rel_matrix_2

class EntropyTest(unittest.TestCase):

	'''
	Tester for the Entropy class.
	'''

	def setUp(self) -> None:
		self.evaluator = Entropy()


	def test_user_exposure(self) -> None:
		top_n = first_example.merge(user_groups, on='user')
		entropy = self.evaluator.evaluate(top_n)
		self.assertAlmostEqual(entropy, 0.91830, delta=1e-5)


	def test_item_exposure(self) -> None:
		top_n = first_example.merge(item_groups, on='item')
		entropy = self.evaluator.evaluate(top_n)
		self.assertAlmostEqual(entropy, 1.48547, delta=1e-5)


class KullbackLeiblerTest(unittest.TestCase):

	'''
	Tester for the KullbackLeibler class.
	'''

	def setUp(self) -> None:
		self.evaluator = KullbackLeibler()


	def test_user_effectiveness(self) -> None:
		top_n = second_example.merge(user_groups, on='user')
		rel_matrix = rel_matrix_1.merge(user_groups, on='user')
		target_representation = pd.DataFrame([['1', 0.5], ['2', 0.5]], columns=['group', 'target_representation'])
		divergence = self.evaluator.evaluate(top_n, target_representation, rel_matrix)
		self.assertAlmostEqual(divergence, 0.08530, delta=1e-5)


	def test_item_exposure(self) -> None:
		top_n = first_example.merge(item_groups, on='item')
		target_representation = pd.DataFrame([['1', 0.2], ['2', 0.3], ['3', 0.5]], columns=['group', 'target_representation'])
		divergence = self.evaluator.evaluate(top_n, target_representation)
		self.assertAlmostEqual(divergence, 0, delta=1e-5)


class MutualInformationTest(unittest.TestCase):

	'''
	Tester for the KullbackLeibler class.
	'''

	def setUp(self) -> None:
		self.evaluator = MutualInformation()


	def test_user_exposure(self) -> None:
		top_n = first_example.merge(user_groups, on='user')
		mi = self.evaluator.evaluate(top_n, 'user')
		self.assertAlmostEqual(mi, 0.25582, delta=1e-5)


	def test_item_exposure(self) -> None:
		top_n = first_example.merge(item_groups, on='item')
		mi = self.evaluator.evaluate(top_n, 'item')
		self.assertAlmostEqual(mi, 0.10570, delta=1e-5)


class CoverageTest(unittest.TestCase):

	'''
	Tester for the Coverage class.
	'''

	def setUp(self) -> None:
		self.evaluator = Coverage()


	def test_coverage_one(self) -> None:
		top_n = first_example[(first_example['item'] != '3') & (first_example['item'] != '4') & (first_example['item'] != '1')]
		cov = self.evaluator.evaluate(top_n, item_groups.item.tolist())
		self.assertAlmostEqual(cov, 0.7)


	def test_coverage_two(self) -> None:
		top_n = first_example[(first_example['item'] != '1') & (first_example['item'] != '2') & (first_example['item'] != '4') & (first_example['item'] != '7') & (first_example['item'] != '9') & (first_example['item'] != '10')]
		cov = self.evaluator.evaluate(top_n, item_groups.item.tolist())
		self.assertAlmostEqual(cov, 0.4)


class NoveltyTest(unittest.TestCase):

	'''
	Tester for the Novelty class.
	'''

	def setUp(self):
		self.evaluator = Novelty()
		novelty = np.vectorize(lambda x: np.mean(-np.log2(int(x))))
		self.novelty = lambda list: np.mean(novelty(list))


	def test_novelty_one(self) -> None:
		top_n = first_example.merge(item_groups, on='item')
		nov = self.evaluator.evaluate(top_n)
		self.assertAlmostEqual(nov, self.novelty(top_n['group'].to_numpy()), delta=1e-5)


	def test_novelty_two(self) -> None:
		top_n = second_example.merge(item_groups, on='item')
		nov = self.evaluator.evaluate(top_n)
		self.assertAlmostEqual(nov, self.novelty(top_n['group'].to_numpy()), delta=1e-5)


if __name__ == '__main__':
	unittest.main()



import numpy as np
import pandas as pd
from abc import ABC, abstractmethod
from recsyslearn.errors import FlagNotValidException
from recsyslearn.utils import eff_matrix, exp_matrix, test_pattern, prob_matrix


class Metric(ABC):

	'''
	Abstract Class for Metrics.
	'''

	def __init__(self) -> None:
		return


	@abstractmethod
	def evaluate(self, top_n: pd.DataFrame) -> float:
		pass


class Coverage(Metric):

	'''
	Coverage evaluator for recommender systems.
	Used formula can be found here https://doi.org/10.1007/s13735-018-0154-2
	'''

	def evaluate(self, top_n: pd.DataFrame, items: list) -> float:

		'''
		Compute the coverage of a model by using its recommendation list.


		Parameters
		----------
		top_n : pd.DataFrame
			Top-N recommendations' lists for every user with items or users already segmented.

		items : list or array-like
			List of items in the dataset.


		Raises
		------
		ColumnsNotMatchException
			If top_n not in the form ('user', 'item', 'rank', 'group').


		Return
		------
		The computed coverage.
		'''

		test_pattern(top_n, ['user', 'item', 'rank'])
		return len(top_n.item.unique().tolist()) / len(items)


class Novelty(Metric):

	'''
	Novelty evaluator for recommender systems.
	Used formula can be found here https://doi.org/10.1007/s13735-018-0154-2
	'''


	def evaluate(self, top_n: pd.DataFrame) -> float:

		'''
		Compute the novelty of a model by using its recommendation list and the segmented item groups.


		Parameters
		----------
		top_n : pd.DataFrame
			Top-N recommendations' lists for every user with items or users already segmented.


		Raises
		------
		ColumnsNotMatchException
			If top_n not in the form ('user', 'item', 'rank', 'group') or if item_group has values not in (0, 1, 2).


		Return
		------
		The computed novelty.
		'''

		test_pattern(top_n, ['user', 'item', 'rank', 'group'])


		top_n.loc[:, 'group'] = pd.to_numeric(top_n.loc[:,'group'])
		top_n = top_n.groupby('user')['group'].apply(lambda x: np.mean(- np.log2(x)))
		return top_n.mean()


class Entropy(Metric):

	'''
	Entropy evaluator for recommender systems.
	'''


	def evaluate(self, top_n: pd.DataFrame, rel_matrix: pd.DataFrame = None) -> float:

		'''
		Compute the entropy of a model by using its recommendation list.


		Parameters
		----------
		top_n : pd.DataFrame
			Top N recommendations' lists for every user with items or users already segmented.

		rel_matrix : pd.DataFrame, default None
			Relevant items for users. It could be, for example, the items with a rating >= threshold.


		Raises
		------
		ColumnsNotMatchException
			If top_n not in the form ('user', 'item', 'rank', 'group').


		Return
		------
		The computed entropy.
		'''

		test_pattern(top_n, ['user', 'item', 'rank', 'group'])


		if rel_matrix is not None:
			top_n = eff_matrix(top_n, rel_matrix)

		top_n = prob_matrix(top_n)

		top_n = top_n[['group', 'rank']].groupby('group', as_index=False).sum()
		top_n['rank'] = top_n['rank'] * np.log2(top_n['rank'])
		return - top_n['rank'].sum()


class KullbackLeibler(Metric):

	'''
	Kullback-Leibler divergence evaluator for recommender systems.
	'''

	def evaluate(self, top_n: pd.DataFrame, target_representation: pd.DataFrame, rel_matrix: pd.DataFrame = None) -> float:

		'''
		Compute the Kullback-Leibler divergence of a model, for a given target representation, by using its recommendation list.


		Parameters
		----------
		top_n : pd.DataFrame
			Top N recommendations' lists for every user with items or users already segmented.

		target_representation : pd.DataFrame
			The target representation desired for each group.

		rel_matrix : pd.DataFrame, default None
			Relevant items for users. It could be, for example, the items with a rating >= threshold.


		Raises
		------
		ColumnsNotMatchException
			If top_n not in the form ('user', 'item', 'rank', 'group') or if target_representation not in the form ('group', 'target_representation').


		Return
		------
		The computed KL Divergence for the given target representation.
		'''

		test_pattern(top_n, ['user', 'item', 'rank', 'group'])
		test_pattern(target_representation, ['group', 'target_representation'])


		if rel_matrix is not None:
			top_n = eff_matrix(top_n, rel_matrix)
		else:
			top_n = exp_matrix(top_n)

		top_n = prob_matrix(top_n)
		top_n = top_n[['group', 'rank']].groupby('group', as_index=False).sum()
		top_n = top_n.merge(target_representation, on='group')
		top_n['rank'] = top_n['rank'] * np.log2(top_n['rank'] / top_n['target_representation'])
		return top_n['rank'].sum()


class MutualInformation(Metric):

	'''
	Mutual Information evaluator for recommender systems.
	'''


	def evaluate(self, top_n: pd.DataFrame, flag: str, rel_matrix: pd.DataFrame = None) -> float:

		'''
		Compute the Mutual Information of a model by using its recommendation list.


		Parameters
		----------
		top_n : pd.DataFrame
			Top N recommendations' lists for every user with items or users already segmented.

		flag : str
			Which actor of the recommendation scenario has been segmented (i.e. user).

		rel_matrix : pd.DataFrame, default None
			Relevant items for users. It could be, for example, the items with a rating >= threshold.


		Raises
		------
		ColumnsNotMatchException
			If top_n not in the form ('user', 'item', 'rank', 'group').

		FlagNotValidException
			If flag is not valid.


		Return
		------
		The computed Mutual Information.
		'''

		test_pattern(top_n, ['user', 'item', 'rank', 'group'])

		not_flagged = {'user': 'item', 'item': 'user'}

		if flag not in list(not_flagged.keys()):
			raise FlagNotValidException()


		if rel_matrix is not None:
			top_n = eff_matrix(top_n, rel_matrix)
		else:
			top_n = exp_matrix(top_n)

		top_n = prob_matrix(top_n)

		P_xy = top_n[[not_flagged[flag], 'group', 'rank']].groupby([not_flagged[flag], 'group'], as_index=False).sum()
		P_xP_y = top_n[[not_flagged[flag], 'group', 'rank']].groupby(not_flagged[flag], as_index=False).sum()
		P_xP_y = P_xy[[not_flagged[flag], 'group']].merge(P_xP_y, on=not_flagged[flag])
		tmp = top_n[['group', 'rank']].groupby('group', as_index=False).sum()
		P_xP_y = P_xP_y.merge(tmp, on=['group'])
		P_xP_y['rank'] = P_xP_y['rank_x'] * P_xP_y['rank_y']
		tmp = P_xP_y[[not_flagged[flag], 'group', 'rank']].merge(P_xy, on=[not_flagged[flag], 'group'])
		tmp['rank'] = tmp['rank_y'] * np.log2(tmp['rank_y'] / tmp['rank_x'])
		return tmp['rank'].sum()



import pandas as pd
import numpy as np

from recsyslearn.errors import ColumnsNotMatchException


def test_pattern(input: pd.DataFrame, pattern: list) -> None:

    '''
    Raise ColumnsNotMatchException if pd.Dataframe header is not as expected.

    Parameters
    ----------
    input : pd.DataFrame
      Input that should be tested.

    pattern : list
      pd.DataFrame column pattern that should be matched. (a header).


    Raise
    ------
    ColumnsNotMatchException
      If input as not expected.

    
    Return
    ------
    Converted columns within the same DataFrame.
    '''

    if list(input.columns) != pattern:
        raise ColumnsNotMatchException(pattern)

    dtypes = {
      'user': str,
      'item': str,
      'rank': float,
      'group': str,
      'target_representation': float
      }

    return input.astype({col: dtypes[col] for col in input.columns if col in dtypes.keys()})
    


def exp_matrix(top_n: pd.DataFrame) -> pd.DataFrame:

    '''
    Compute exposure matrix for given recommendation lists.

    Parameters
    ----------
    top_n : pd.DataFrame
      	Recommendation lists per user in the form (user, item, rank, group).


    Raise
    -----
    ColumnsNotMatchException
      	If top_n is not in the from (user, item, rank, group).


    Return
    ------
    top_n : pd.DataFrame
      	The DataFrame with computed exposure.
    '''

    test_pattern(top_n, ['user', 'item', 'rank', 'group'])

    top_n['rank'] = 1 / np.log2(1 + top_n['rank'])
    return top_n


def prob_matrix(top_n: pd.DataFrame) -> pd.DataFrame:

    '''
    Compute probability distribution matrix for given recommendation lists.

    Parameters
    ----------
    top_n : pd.DataFrame
      	Recommendation lists per user in the form (user, item, rank, group).


    Return
    ------
    top_n : pd.DataFrame
      	The DataFrame with computed probability distribution.


    Raise
    -----
    ColumnsNotMatchException
      	If top_n is not in the from (user, item, rank, group).
    '''

    test_pattern(top_n, ['user', 'item', 'rank', 'group'])
    
    top_n['rank'] = top_n['rank'] / top_n['rank'].sum()
    return top_n


def eff_matrix(top_n: pd.DataFrame, rel_matrix: pd.DataFrame) -> pd.DataFrame:

    '''
    Compute effectiveness matrix for given recommendation lists.

    Parameters
    ----------
    top_n : pd.DataFrame
      	Recommendation lists per user in the form (user, item, rank, group).


    Return
    ------
    top_n : pd.DataFrame
      	The DataFrame with computed effectiveness.


    Raise
    -----
    ColumnsNotMatchException
      	If top_n header is not in the from (user, item, rank, group).
      	If rel_matrix header is not in the form (user, item, rank).
    '''

    test_pattern(top_n, ['user', 'item', 'rank', 'group'])
    test_pattern(rel_matrix, ['user', 'item', 'rank', 'group'])

    top_n = exp_matrix(top_n)
    top_n = top_n.merge(rel_matrix, on=['user', 'item', 'group'], how='outer')
    top_n.loc[:, ['rank_x', 'rank_y']] = top_n.loc[:, ['rank_x', 'rank_y']].fillna(0)
    top_n['rank'] = top_n['rank_x'] * top_n['rank_y']
    return top_n[['user', 'item', 'rank', 'group']]



import unittest
from recsyslearn.segmentations import ActivitySegmentation, InteractionSegmentation
import pandas as pd
from tests.utils import dataset_item_example, dataset_user_example

class InteractionSegmentationTest(unittest.TestCase):

	'''
	Tester for the InteractionSegmentation class.
	'''

	def setUp(self) -> None:
		self.interactionSegmentator = InteractionSegmentation()

	def test_segmentation(self) -> None:
		segmented_groups = self.interactionSegmentator.segment(dataset_item_example)
		self.assertTrue(segmented_groups.loc[segmented_groups['item'] == '1', 'group'].eq('1').all())
		self.assertTrue(segmented_groups.loc[segmented_groups['item'] == '2', 'group'].eq('2').all())


class ActivitySegmentationTest(unittest.TestCase):

	'''
	Test for the ActivitySegmentation class.
	'''

	def setUp(self) -> None:
		self.userSegmentator = ActivitySegmentation()

	def test_segmentation(self) -> None:
		segmented_groups = self.userSegmentator.segment(dataset_user_example)
		self.assertTrue(segmented_groups.loc[segmented_groups['user'] == '1', 'group'].eq('1').all())
		self.assertTrue(segmented_groups.loc[segmented_groups['user'] == '2', 'group'].eq('2').all())
		self.assertTrue(segmented_groups.loc[segmented_groups['user'] == '3', 'group'].eq('2').all())


if __name__ == '__main__':
	unittest.main()



class ColumnsNotMatchException(Exception):
    def __init__(self, pattern, message='Columns doesn\'t match the pattern.') -> None:
        super().__init__(f'{message} {pattern}')

class SegmentationNotSupportedException(Exception):
    def __init__(self, message='Number of supported group is between 1 and 3.') -> None:
        super().__init__(message)

class WrongProportionsException(Exception):
    def __init__(self, message='Proportions doesn\'t cover all the items/users.') -> None:
        super().__init__(message)

class FlagNotValidException(Exception):
    def __init__(self, message='Invalid flag.') -> None:
        super().__init__(message)



#!/usr/bin/env python

'The setup script.'

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['numpy', 'pandas']

test_requirements = ['numpy', 'pandas']

setup(
    author='Giulio Davide Carparelli',
    author_email='giulio.davide.97@gmail.com',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
		'Topic :: Scientific/Engineering :: Mathematics',
		'Topic :: Software Development :: Version Control :: Git',
		'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.8',
    ],
    description='A small library to compute fairness of recommender systems.',
    install_requires=requirements,
    license='GNU General Public License v3',
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='recsyslearn',
    name='recsyslearn',
    packages=find_packages(include=['recsyslearn', 'recsyslearn.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/giuliowaitforitdavide/recsyslearn',
    version='0.4.1',
    zip_safe=False,
)



import numpy as np
import pandas as pd
from abc import ABC, abstractmethod

from recsyslearn.errors import SegmentationNotSupportedException, WrongProportionsException


class Segmentation(ABC):

	'''
	Abstract class for different group segmentations.
	'''

	def __init__(self) -> None:
		super().__init__()

	
	@abstractmethod
	def segment(self, dataset: pd.DataFrame) -> None:
		pass


class InteractionSegmentation(Segmentation):

	'''
	Segmentation of items based on the number of interaction they have.
	'''

	def segment(self, dataset: pd.DataFrame, proportion: list = [0.8, 0.2], min_interaction: int = 0) -> pd.DataFrame:

		'''
		Segmentation of items based on their interactions with different users.


		Parameters
		----------
		dataset : pd.DataFrame
			The complete dataset.

		proportion : list, default [0.8, 0.2]
			The proportion of interactions wanted for every group. Its length should be between 1 and 3.


		Raises
		------
		SegmentationNotSupportedException
			If len(proportion) not in (1, 2, 3).
		
		WrongProportionsException
			If sum(proportion) is not 1, which means it doesn't cover all the items/users.

		
		Return
		------
		DataFrame with items and belonging group.
		'''

		if len(proportion) not in (1, 2, 3):
			raise SegmentationNotSupportedException()

		if sum(proportion) != 1:
			raise WrongProportionsException()

		if len(proportion) == 1:
			return dataset

		item_groups = dataset.groupby('item').size().reset_index(name='count').sort_values('count', ascending=False) 

		tmp = item_groups[item_groups['count'] > min_interaction]
		n_int = tmp['count'].sum()

		short_thr = np.rint(proportion[0] * n_int)
		mid_thr = np.rint(proportion[1] * n_int) + short_thr
		tmp.loc[:, 'cumulative_sum'] = tmp['count'].cumsum()
		
		short_head = tmp.loc[tmp['cumulative_sum'].lt(short_thr), 'item']
		mid_tail = tmp.loc[tmp['cumulative_sum'].lt(mid_thr) & ~tmp['item'].isin(short_head), 'item']
		conditions = [item_groups['item'].isin(short_head), item_groups['item'].isin(mid_tail)]
		choices = ['1', '2']
		item_groups.loc[:, 'group'] = np.select(conditions, choices, default='3' if len(proportion) == 3 else '2')
		return item_groups[['item', 'group']]


class ActivitySegmentation(Segmentation):

	'''
	Segmentation of users based on their number of interaction.
	'''

	def segment(self, dataset: pd.DataFrame, proportion: list = [0.1, 0.9], min_interaction: int = 0) -> pd.DataFrame:

		'''
		Segmentation of users based on their interactions with different items.


		Parameters
		----------
		dataset : pd.DataFrame
			The complete dataset.

		proportion : list, default [0.8, 0.2]
			The proportion of interactions wanted for every group.


		Raises
		------
		SegmentationNotSupportedException
			If len(proportion) not in (1, 2, 3).
		
		WrongProportionsException
			If sum(proportion) is not 1, which means it doesn't cover all the items/users.

		
		Return
		------
		DataFrame with users and belonging group.
		'''

		if len(proportion) not in (1, 2, 3):
			raise SegmentationNotSupportedException()

		if sum(proportion) != 1:
			raise WrongProportionsException()

		if len(proportion) == 1:
			return dataset

		user_groups = dataset.groupby('user').size().reset_index(name='count')
		user_groups = user_groups[user_groups['count'] > min_interaction]
		
		# Adding some random interactions to remove ties
		user_groups.loc[:, 'count'] = user_groups.loc[:, 'count'].apply(lambda x: x + np.random.choice(list(range(10))))
		user_groups = user_groups.sort_values('count', ascending=False)
		user_groups.loc[:, 'count'] = np.arange(user_groups.shape[0]) + 1
		threshold = round(user_groups.shape[0] * proportion[0]) if round(user_groups.shape[0] * proportion[0]) > 0 else 1

		user_groups.loc[user_groups['count'] <= threshold, 'group'] = '1'
		user_groups.loc[user_groups['count'] > threshold, 'group'] = '2'

		return user_groups[['user', 'group']]
